/* 
 * $Author: djh $ $Date: 1993/11/29 06:48:59 $
 * $Header: /mac/src/cap60/lib/cap/RCS/abauxddp.c,v 2.2 1993/11/29 06:48:59 djh Rel djh $
 * $Revision: 2.2 $
 *
 */

/*
 * abauxddp.c - Datagram Delivery Protocol for native appletalk under A/UX
 *
 * AppleTalk package for UNIX (4.2 BSD).
 *
 * Copyright (c) 1986, 1987, 1988 by The Trustees of Columbia University
 *  in the City of New York.
 *
 *
 * Edit History:
 *
 *  June 19, 1986    Schilit    Created.
 *  July  9, 1986    CCKim      Clean up some of Bill's stuff and allow
 *                              Appletalk protocols on ethernet with CAP
 *  Feb. 1988 Charlie - don't like the way the encapsulation code runs
 *   into the DDP code.  Drop out all the encapsulation code into
 *   another module (interface dep) and drop out part of DDP into it.
 *
 *  1990   William Roberts	Add support for A/UX native appletalk
 *
 *  1995   Chris Benndorf   Changed to plug CAP into NetAtalk
 *         (chrisb@ganzfix.duew.eunet.de)
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>

/* Use the direct Apple support for ddp. We insist on a long form
 * DDP header, and we can open our own file descriptor instead of
 * calling abOpen to do it for us. iDDPOpenSocketIOV and DDPClose
 * are modified to set up the fdlistener stuff that abOpen/abClose
 * used to deal with.
#include <at/appletalk.h>
#include <at/lap.h>
#include <at/ddp.h>
 */

#include "ablnx.h"

/* #include <sys/stropts.h>        /* for strioctls */
#include <netinet/in.h>         /* for ntohs */

/* Fortunately the CAP include and the Apple ones don't conflict */

#include <netat/appletalk.h>    /* for private, boolean etc */
#include <linux/atalk.h>    /* for private, boolean etc */
#include "cap_conf.h"

#ifndef DONT_DOCHKSUM
# define DO_CHKSUM 1
#else
# define DO_CHKSUM 0
#endif
boolean dochecksum = DO_CHKSUM;	/* can be patched if necessary */

short lap_proto = LAP_KERNEL;   /* kernel appletalk support */

extern int errno;

/* room for up to ddp + 1 for data buffer */
#define DDPIOVLEN (IOV_DDP_SIZE+1)
private DDP ddph;
private byte ddpdatabuffer[ddpMaxData];
private struct iovec rddpiov[DDPIOVLEN] = {
  {NULL, 0},                    /* LAP header */
  {(caddr_t)&ddph, ddpSize},    /* DDP header (me) (redundant) */
  {(caddr_t)ddpdatabuffer, ddpMaxData} /* ddp data */
};

typedef struct {
  int (*lproc)();               /* socket listener routine */
  int flags;                    /* flags */
#define DDPL_OLDINTERFACE 1
  int fd;
  struct sockaddr_at at_sock;           /* DDP file descriptor */
} LISENTRY;                     /* listener entry */

private LISENTRY ddpl[ddpMaxSkt+1]; /* table of listeners */


/* abInit(disp)
 *
 * Get things started and find out our Node and Net
 * Maybe we should wait until later in case a seed bridge starts up?
 */

export word this_net;
export byte this_node;
export byte this_intfno;
export word bridge_net;         /* do we care about this any more? */
export byte bridge_node;
export word nis_net;            /* do we care about this any more? */
export byte nis_node;
export char this_zone[34];
export char async_zone[34];
export char this_intf[50];

export DBUG dbug;

ddp_open( port )
  at_socket *port;
{
    int s,nlen;
    struct sockaddr_at new_sock;

    if (( s = socket( AF_APPLETALK, SOCK_DGRAM, 0 )) < 0 ) {
        printf("Cannot open socket\n");
        return -1;
    }

    bzero( (char *) &new_sock, sizeof( struct sockaddr_at ));
    new_sock.sat_family = AF_APPLETALK;
    new_sock.sat_addr.s_net = ATADDR_ANYNET;
    new_sock.sat_addr.s_node = ATADDR_ANYNODE;
    new_sock.sat_port = ( *port == 0 ? ATADDR_ANYPORT : *port );

    if ( bind( s, (struct sockaddr *) &new_sock,
            sizeof( struct sockaddr_at )) != 0 ) {
        printf("Cannot bind socket\n");
        return -1;
    }
    /* get the real address from the kernel */
    nlen = sizeof( struct sockaddr_at);
    if ( getsockname( s, (struct sockaddr *) &new_sock, &nlen ) != 0 ) {
        printf("Cannot get socket name\n");
        return -1;
    }
    *port=new_sock.sat_port;
    ddpl[*port].at_sock.sat_family=new_sock.sat_family;
    ddpl[*port].at_sock.sat_addr.s_net=new_sock.sat_addr.s_net;
    ddpl[*port].at_sock.sat_addr.s_node=new_sock.sat_addr.s_node;
    ddpl[*port].at_sock.sat_port=new_sock.sat_port;

    if (dbug.db_ddp) {
        printf("Opened socket %d at %d.%d, s=%d for PID=%d\n", new_sock.sat_port,s,
                new_sock.sat_addr.s_net, new_sock.sat_addr.s_node, getpid());
    } 

   
    return s;
}

ddp_close ( sock )
    int sock;
{
    if (dbug.db_ddp) {
        printf("Closing socket %d\n", sock);
    }
    
    if ( close( sock ) < 0 ) {
        return -1;
    }
}

abInit(disp)
int disp;
{
    int fd, x, nt1,nd1,nt2,nd2;
    FILE *fp;
    at_ddp_cfg_t cfg;
    at_socket socket;

    init_fdlistening();
    DDPInit();

    /* find out our address */

    socket = 0;         /* don't care */
    fd = ddp_open(&socket);

    if (fd < 0) {
	perror("ddp_open");
	exit(0);
    }

    if (dbug.db_lap) {
		printf("abInit: [ddp %3d.%02d, %d] starting\n",
	    ntohs(ddpl[socket].at_sock.sat_addr.s_net)>>8,
	    ntohs(ddpl[socket].at_sock.sat_addr.s_net) &0x0FF,
	    ddpl[socket].at_sock.sat_addr.s_node);
    }

	this_net  = ddpl[socket].at_sock.sat_addr.s_net;
    this_node = ddpl[socket].at_sock.sat_addr.s_node;
   	bridge_net  = ddpl[socket].at_sock.sat_addr.s_net;
   	bridge_node = ddpl[socket].at_sock.sat_addr.s_node;
    
    ddpl[socket].lproc = NILPROC;
    ddpl[socket].flags = 0;
    ddpl[socket].at_sock.sat_family=-1;
    ddpl[socket].at_sock.sat_addr.s_net=-1;
    ddpl[socket].at_sock.sat_addr.s_node=-1;
    ddpl[socket].at_sock.sat_port=-1;

    x = ddp_close(fd);
    if (x < 0) {
	perror("ddp_close");
	exit(0);
    }
}

void
zoneset(zonename)
char *zonename;
{
    /* ignored */
}

OSErr
GetNodeAddress(mynode, mynet)
int *mynode, *mynet;
{
    *mynode = this_node;
    *mynet  = this_net;
    return (noErr);
}

/*
 * add these for compatibility with other changes ...
 *
 */

GetMyAddr(addr)
AddrBlock *addr;
{
    addr->net = this_net;
    addr->node = this_node;
}

SetMyAddr(addr)
AddrBlock *addr;
{
    this_net = addr->net;
    this_node = addr->node;
}

GetNisAddr(addr)
AddrBlock *addr;
{
    addr->net = nis_net;
    addr->node = nis_node;
}

SetNisAddr(addr)
AddrBlock *addr;
{
    nis_net = addr->net;
    nis_node = addr->node;
}



DDPInit()
{
  int i;

  /* initialize ddp listener array */
  for (i = 0; i < ddpMaxSkt+1; i++) {
    ddpl[i].lproc = NILPROC;
    ddpl[i].flags = 0;
    ddpl[i].at_sock.sat_family=-1;
    ddpl[i].at_sock.sat_addr.s_net=-1;
    ddpl[i].at_sock.sat_addr.s_node=-1;
    ddpl[i].at_sock.sat_port=-1;
  }

}


/*
 * OSErr DDPOpenSocket(int *skt, ProcPtr sktlis)
 * OSErr DDPOpenSocketIOV(int *skt,ProcPtr sktlis,struct iovec *iov,int iovlen)
 *
 * Open a DDP socket and optionally install a socket listener to the
 * listener table.  If skt is nonzero (it must be in the range of
 * 1 to 127) it specifies the socket's number.  If skt is 0 then
 * DDPOpenSocket dynamically assigns a socket number in the range
 * 128 to 254, and returns it in skt.  You can actually specify a socket
 * in the 128 to 254 range, but it's not a good idea :-).
 *
 * sktlis contains a pointer (ProcPtr) to the socket listener; if
 * it is NILPROC, the default listener will be used (NYI)
 *
 * If calling DDPOpenSocketIOV, then the iovector passed must be of
 * size (IOV_DDP_SIZE+1) (length for ddp+lap) plus one (data for caller).
 * In addition, it must be filled in from DDP_LVL+1 to the end
 *
 *
 * The listener is defined as:
 *  XXX_listener(int skt, PKT *pkt, int len, AddrBlock *addr)
 * if called from DDPOpenSocket and:
 *  XXX_listener(int skt, struct iovec *iov, int iovlen, AddrBlock *addr)
 * if called from DDPOpenSocketIOV
 *
 * The iov passed back to the listener will start after the ddp header
 * block
 *
*/
OSErr
DDPOpenSocketIOV(skt, sktlis, iov, iovlen)
int *skt;
ProcPtr sktlis;
struct iovec *iov;
int iovlen;
{
  int fd;
  at_socket s;
  int defDDPlis();      /* which always fails anyway - clearly a throwback */
  int ddp_upcall();

  /* allow 0 - means dynamic assignment */
  s = *skt;                     /* socket wanted */
  if (s >= ddpMaxSkt) {
    fprintf(stderr,"ddpOpenSocket: skt out of range\n");
    exit(0);
  }
  /* open the socket please */
  if ((fd = ddp_open(&s)) < 0) {
    if (dbug.db_ddp)
      fprintf(stderr,"ddp: open socket - socket open failed: %d\n", errno);
    return(fd);              /* return error if failed */
  }
  *skt = s;             /* socket number actually obtained */
  if (dbug.db_ddp) {
    fprintf(stderr, "ddp: open socket: opened socket %d\n", s);
  }

  iov[IOV_DDP_LVL].iov_base = (caddr_t)&ddph; /* install */
  iov[IOV_DDP_LVL].iov_len = ddpSize; /* install */

  /* add default or user's listener */
  ddpl[s].lproc = ((sktlis == NILPROC) ? defDDPlis : sktlis);
  ddpl[s].flags = 0;
  ddpl[s].fd = fd;

  /* add the file descriptor to the list of those listened for */
  fdlistener(fd, ddp_upcall, iov+IOV_DDP_LVL, iovlen-IOV_DDP_LVL);

  return(noErr);                        /* and return */
}


OSErr
DDPOpenSocket(skt,sktlis)
int *skt;
ProcPtr sktlis;
{
  OSErr err;

  err = DDPOpenSocketIOV(skt, sktlis, rddpiov, DDPIOVLEN);
  if (err == noErr) {
    ddpl[*skt].flags = DDPL_OLDINTERFACE;
  }
  return(err);
}


/*
 * OSErr DDPCloseSocket(int skt)
 *
 * DDPCloseSocket closes the skt, cancels all pending DDPRead calls
 * that have been made on that socket, and removes the socket listener
 * procedure.
 *
*/

OSErr
DDPCloseSocket(skt)
int skt;
{
  int err;

  if (skt == 0 || skt >= ddpMaxSkt) {
    if (dbug.db_ddp)
      fprintf(stderr, "ddpCloseSocket: Socket %d out of range\n", skt);
    return;
  }
  ddpl[skt].lproc = NILPROC;    /* no procedure */

  fdunlisten(ddpl[skt].fd);     /* stop listening to the socket */

  err=ddp_close(ddpl[skt].fd);
  ddpl[skt].fd = -1;
  ddpl[skt].lproc = NILPROC;
  ddpl[skt].flags = 0;
  ddpl[skt].at_sock.sat_family=-1;
  ddpl[skt].at_sock.sat_addr.s_net=-1;
  ddpl[skt].at_sock.sat_addr.s_node=-1;
  ddpl[skt].at_sock.sat_port=-1;
  return(err);
}

OSErr
DDPWrite(abr)
abRecPtr abr;
{
  struct iovec iov[IOV_DDP_SIZE+1];

  iov[IOV_DDP_LVL+1].iov_base = (caddr_t) abr->proto.ddp.ddpDataPtr;
  iov[IOV_DDP_LVL+1].iov_len = abr->proto.ddp.ddpReqCount;

  return(DDPWriteIOV(abr,iov,IOV_DDP_SIZE+1));
}

/*
 * DDPWriteIOV
 *
 * DDPWriteIOV builds up DDP header and then passes off to routeddp
 *  who decides where to send it.  In the most cases, we'll probably
 *  have a version of routeddp per "network" type so we can "optimize"
 *
*/
/*ARGSUSED*/
OSErr
DDPWriteIOV(abr,iov,iovl)
abRecPtr abr;
struct iovec iov[];
int iovl;
{
  struct sockaddr_at      target;
  at_ddp_t ddp;
  ddpProto *dpr;
  at_socket skt;
  char     *p,buf[1024];
  int i,err;

  dpr = &abr->proto.ddp;
  skt = dpr->ddpSocket; /* our socket number */

  if (skt == 0 || skt >= ddpMaxSkt || ddpl[skt].lproc == NILPROC) {
	return -1;
  }

  /* the DDP streams driver fills in everything else given
   * just checksum, dst_net, dst_node, dst_socket, type and data.
   * The length is inferred from the write or writev command.
   */
  p = buf;
  *p = dpr->ddpType;
  p++;

  
  for (i=IOV_DDP_LVL+1; i<iovl; i++) {
    bcopy( iov[i].iov_base, p, iov[i].iov_len );
    p+=iov[i].iov_len;
  }

  bzero( &target, sizeof( struct sockaddr_at ));
  target.sat_family       = AF_APPLETALK;
  target.sat_addr.s_net  = dpr->ddpAddress.net;
  target.sat_addr.s_node = dpr->ddpAddress.node;
  target.sat_port         = dpr->ddpAddress.skt;

/*
for (i=IOV_DDP_LVL; i<iovl; i++) logit(0,"IOVLEN[%d]:%d",i,iov[i].iov_len);
logit(0,"BUFFLEN:%d",p - buf);
*/

  err=sendto( ddpl[skt].fd, buf , p - buf , 0, &target,
          sizeof( struct sockaddr_at ));

  /* The DDP streams driver deals with LAP headers etc, so must
   * adjust the iov to point to the start of the DDP stuff
   */
  iov += IOV_DDP_LVL; iovl -= IOV_DDP_LVL;

  iov->iov_base = (caddr_t) &ddp; 
  iov->iov_len = ddpSize;


  if (err <= 0) {
    if (dbug.db_ddp) {
	perror("ddp_write failed");
    }
    return -1;
  }
  return 0;
}

/*ARGSUSED*/
OSErr
DDPRead(abr,retCkSumErrs,async)
abRecPtr abr;
int retCkSumErrs,async;
{
  fprintf(stderr,"DDPRead NYI\n");
}

OSErr
DDPRdCancel(abr)
abRecPtr abr;
{
  fprintf(stderr,"DDPRdCancel NYI\n");
}

defDDPlis(skt,ddp,len,addr)
DDP *ddp;
AddrBlock *addr;
{
  fprintf(stderr,"defDDPlis NYI\n");
/***** copy data into user buffer *****/
}

/*
 * ddp_protocol(ddp,len)
 *
 * ddp_protocol is the installed LAP protocol handler for protocol
 * type lapDDP (2).  This routine gets called by LAP when a packet
 * is received with lapDDP in the protocol field.  ddp_protocol
 * passes the packet to the socket listener installed by the
 * DDPOpenSocket call.
 *
 * With direct support for DDP, this routine is entered from the
 * very simple ddp_upcall routine installed as the fdlistener for
 * the appropriate file descriptor.
 *
 * Caller provides an iov pointing to DDP header
*/

int ddp_protocol(iov, iovlen, plen)
struct iovec *iov;
int iovlen;
int plen;
{
  at_socket skt;
  AddrBlock addr;
  int len, cnt, oldstyle;
  at_ddp_t *ddp;

  /* iovlen == 1 means just a ddp header */
  if (iovlen < 1 || iov->iov_len != ddpSize) {
    return -1;
  }


  ddp = (at_ddp_t *)iov->iov_base; /* it must be aligned ok already*/
  len = ddp->length - ddpSize ;
  if (plen < len) {             /* not enought data? */
    if (dbug.db_ddp)
      fprintf(stderr, "BAD PACKET: ddp reports more data than came in\n");
    return;                     /* drop pkt */
  }

  skt = ddp->dst_socket;

  if (ddpl[skt].lproc == NILPROC) { /* listener proc for socket */
    if (dbug.db_ddp)
      fprintf(stderr,"ddp_protocol: no socket listener!\n");
    return;
  }

  addr.net = ddp->src_net;
  addr.node = ddp->src_node;
  addr.skt = ddp->src_socket;
  iov++;                        /* skip ddp header */
  iovlen--;
  plen -= ddpSize;              /* reduce to data size */
  if (iovlen < 1) {
    return 0;   /* nothing to send, so drop it */
  }
  if (ddpl[skt].lproc) {        /* be postitive */
    if (ddpl[skt].flags & DDPL_OLDINTERFACE)
      (*ddpl[skt].lproc)(skt,ddp->type,iov->iov_base,plen,&addr);
    else
      (*ddpl[skt].lproc)(skt, ddp->type, iov, iovlen, plen, &addr);
  }
  return 0;
}

/*private*/ int ddp_upcall(fd, iov, iovlen)
int fd;
struct iovec *iov;
int iovlen;
{
    char     *p, buf[1024];
    struct sockaddr_at sat;
    int i,nbytes,satlen,skt;
    at_ddp_t *ddp;
    /* struct iovec *iov; */
    
    skt=fd2ddp_skt(fd);
    
    bzero( &sat, sizeof( struct sockaddr_at ));
    p=ddp=buf;

    satlen=sizeof( struct sockaddr_at  );
    nbytes = recvfrom(fd, &ddp->type , 1024 - ddpSize, 0, &sat, &satlen);

    nbytes+= ddpSize - 1;

    ddp->unused= 0;
    ddp->hopcount= 0;
    ddp->length= nbytes + ddpSize - 1;  /* -1 for ddp.type */
    ddp->checksum= 0 ;
    ddp->dst_net= ddpl[skt].at_sock.sat_addr.s_net;
    ddp->src_net= sat.sat_addr.s_net;
    ddp->dst_node= ddpl[skt].at_sock.sat_addr.s_node;
    ddp->src_node= sat.sat_addr.s_node;
    ddp->dst_socket= skt;
    ddp->src_socket= sat.sat_port;

    p+=ddpSize;

/* for (i=0; i<iovlen; i++) logit(0,"IOVLEN[%d]:%d",i,iov[i].iov_len); */

    bcopy( ddp , iov->iov_base, ddpSize); /*  */
    iov->iov_len = ddpSize;          


    if ( iovlen == 2 ) {
       bcopy( p , iov[1].iov_base, nbytes - ddpSize ); /*  */
       iov[1].iov_len = nbytes - ddpSize ;          
    } else {
       bcopy( p , iov[1].iov_base, atpSize ); /*  */
       iov[1].iov_len = atpSize ;          
 
       p+=atpSize;

       bcopy( p , iov[2].iov_base, nbytes - atpSize - ddpSize );  /* */
       iov[2].iov_len =  nbytes - atpSize - ddpSize;
    }

/* for (i=0; i<iovlen; i++) logit(0,"IOVLEN[%d]:%d",i,iov[i].iov_len); */

    return (ddp_protocol(iov, iovlen, nbytes));
}


/*
 * Call with TRUE or FALSE - FALSE means ignore, TRUE means don't ignore
 *
*/
checksum_error(which)
boolean which;
{
  /* who cares - we certainly don't */
}


/* Utility routine to sprintf an address in a standard format
 *
 *      net_hi.net_lo,node socket
 *
 * It returns its first argument, for convenient use in printf.
 */
char *appletalk_ntoa(buf, addr)
char buf[];
AddrBlock *addr;
{
    sprintf(buf, "%d.%02d,%d %d",
	((addr->net)>>8)& 0x0FF, (addr->net)& 0x0FF,
	addr->node, addr->skt);
    return buf;
}

/* int ddp_skt2fd(skt)
 *
 * Converts a socket number into a file descriptor, or returns -1.
 * Used by upper layers which have converted over to A/UX support,
 * since CAP is otherwise keen only to pass around socket numbers.
 */

int ddp_skt2fd(skt)
byte skt;
{
    if (skt > 0 && skt <= ddpMaxSkt && ddpl[skt].lproc != NILPROC) {
	return ddpl[skt].fd;
    }

    /* invlaid skt or no socket listener - presumed not open */
    return -1;
}

int fd2ddp_skt(fd)
int fd;
{
    int i;
      for (i=0;i<ddpMaxSkt;i++)        if (ddpl[i].fd==fd) return i;
}

int ddp_skt2addr(port,saddr)
struct sockaddr_at *saddr;
byte port;
{
    if (port > 0 && port <= ddpMaxSkt ) {
        saddr->sat_family=ddpl[port].at_sock.sat_family;
        saddr->sat_addr.s_net=ddpl[port].at_sock.sat_addr.s_net;
        saddr->sat_addr.s_node=ddpl[port].at_sock.sat_addr.s_node;
        saddr->sat_port=ddpl[port].at_sock.sat_port;
	return 0;
    }

    /* invlaid skt or no socket listener - presumed not open */
    return -1;
}

/*
 * avoid using global variables
 *
 */

OSErr
GetBridgeAddress(addr)
AddrBlock *addr;
{
  addr->net = bridge_net;
  addr->node = bridge_node;
  return(noErr);
}

OSErr
SetBridgeAddress(addr)
AddrBlock *addr;
{
  bridge_net = addr->net;
  bridge_node = addr->node;
  return(noErr);
}
